""" Define helper methods to reduce the boilerplate code for a server
"""

import logging

from fastapi import Request
from fastapi.exceptions import RequestValidationError
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from styler_rest_framework import message
from styler_rest_framework.config import defaults
from styler_rest_framework.middlewares.fastapi import exception_middleware
from styler_rest_framework.logging.error_reporting import (
    google_error_reporting_handler
)


def api_error_reporting_handler(service=None):  # pragma: no coverage
    try:
        service_name = service or defaults.ERROR_HANDLER_SERVICE
        from google.cloud import error_reporting
        handler = google_error_reporting_handler(service=service_name)

        def error_handler(request, exc):
            http_context = error_reporting.HTTPContext(
                method=request.method, url=request.path)
            handler(http_context=http_context)

        return error_handler
    except Exception:
        logging.warning('Could not find start api error reporting')
        return None


def add_middlewares(
            app,
            error_handler=None,
            service=None,
            handle_exceptions_args=None
        ):
    """ Append default exception middleware

        Args:
            app: FastAPI app
            error_handler: exception handler
            service: service name
            handle_exceptions_args: dict of parameters to the middleware
    """
    handle_exceptions_args = handle_exceptions_args or {}
    error_handler = error_handler or api_error_reporting_handler(
        service=service)

    exception_middleware.add_exception_middleware(
        app, error_handler=error_handler, **handle_exceptions_args)


def setup_validation_handler(
            app,
            validation_error_code=422,
            validation_code='validation_error'):  # pragma: no coverage
    """ Set the validation error by overriding the RequestValidationError
    """
    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(
            request: Request, exc: RequestValidationError):
        locale = request.headers.get('Accept-Language', 'ja')
        errors = {}
        for err in exc.errors():
            field = '.'.join(err['loc'][1:]) or 'error'
            err_type = err['type'].replace('.', '_')
            errors[field] = message.get(
                f'pydantic_validation.{err_type}',
                locale=locale,
                **err.get('ctx', {})
            )

        return JSONResponse(
            status_code=validation_error_code,
            content={'code': validation_code, 'reason': errors},
        )

    def custom_openapi():
        if app.openapi_schema:
            return app.openapi_schema
        openapi_schema = get_openapi(
            title=defaults.SERVICE_NAME,
            version=defaults.VERSION,
            routes=app.routes
        )
        openapi_schema['components']['schemas'].pop('ValidationError')
        openapi_schema['components']['schemas']['HTTPValidationError'] = {
            'title': 'ValidationError',
            'type': 'object',
            'properties': {
                'code': {
                    'type': 'string'
                },
                'reason': {
                    'type': 'object',
                    'additionalProperties': {
                        'type': 'string'
                    }
                }
            },
            'required': ['code', 'reason']
        }
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    app.openapi = custom_openapi



""" Tests for the Migration tools
"""

import pytest
from unittest.mock import Mock, patch

from styler_rest_framework.helpers import migration


class TestCheckAndRetryMigration:
    """ Tests for check_and_retry_migration
    """
    def test_version_up_to_date(self):
        """ It should return without errors
        """
        migration.get_heads_version = Mock(return_value=['head'])
        migration.get_current_version = Mock(return_value='head')

        migration.check_and_retry_migration('mock_path')

        migration.get_heads_version.assert_called_once()
        migration.get_current_version.assert_called_once()

    def test_multiple_head(self):
        """ It should raise migration error
        """
        migration.get_heads_version = Mock(return_value=['head1', 'head2'])
        migration.get_current_version = Mock(return_value='head')

        with pytest.raises(migration.MigrationError) as expected:
            migration.check_and_retry_migration('mock_path')

        assert "multiple heads detected!" in str(expected.value)

        migration.get_heads_version.assert_called_once()

    @patch('alembic.command.upgrade')
    @patch('logging.warning')
    def test_retry_success(self, log_mock, upgrade_mock):
        """ It should return without errors
        """
        migration.get_heads_version = Mock(return_value=['head'])
        migration.get_current_version = Mock(return_value='old_version')

        migration.check_and_retry_migration('mock_path')

        migration.get_heads_version.assert_called_once()
        migration.get_current_version.assert_called_once()
        upgrade_mock.assert_called_once()
        log_mock.assert_any_call('alembic version is not up-to-date, \
        will retry migration...')

    @patch('logging.warning')
    def test_retry_failed(self, log_mock):
        """ It should raise migration error
        """
        migration.get_heads_version = Mock(return_value=['head'])
        migration.get_current_version = Mock(return_value='old_version')

        with patch('alembic.command.upgrade',
                   side_effect=Exception()) as upgrade_mock:
            with pytest.raises(migration.MigrationError) as expected:
                migration.check_and_retry_migration('mock_path',
                                                    max_retry_count=3)

            migration.get_heads_version.assert_called_once()
            migration.get_current_version.assert_called_once()
            assert 3 == upgrade_mock.call_count
            assert "Could not execute alembic migrations." in str(
                expected.value)

    def test_missing_parameter(self):
        with pytest.raises(migration.MigrationError) as expected:
            migration.check_and_retry_migration('')

        assert "parameter required: cfg_path" in str(expected.value)



import logging
import time
from random import randint

from alembic.config import Config
from alembic import command


class MigrationError(Exception):
    """MigrationError class for exceptions in this module."""
    pass


def get_current_version(alembic_cfg):
    captured_text = None

    def print_stdout(text, *arg):
        nonlocal captured_text
        captured_text = text

    alembic_cfg.print_stdout = print_stdout
    command.current(alembic_cfg)
    return captured_text


def get_heads_version(alembic_cfg):
    captured_text = []

    def print_stdout(text, *arg):
        nonlocal captured_text
        captured_text.append(text)
    alembic_cfg.print_stdout = print_stdout
    command.heads(alembic_cfg)
    return captured_text


def retry_migration(alembic_cfg, count, max_count, max_backoff=32):
    try:
        logging.warning(f'retry migration now... (retry count:{count+1})')
        command.upgrade(alembic_cfg, 'head')
        logging.warning('retry migration succeed!')
        return
    except Exception:
        backoff = min(2**(count)+randint(1, 1000)/1000, max_backoff)
        count += 1
        if count >= max_count:
            raise MigrationError('Could not execute alembic migrations.')
        logging.warning(f'retry migration failed, \
            will retry again after {backoff} seconds...')
        time.sleep(backoff)
        retry_migration(alembic_cfg, count, max_count, max_backoff)


def check_and_retry_migration(cfg_path, max_retry_count=3):
    if not cfg_path:
        raise MigrationError('parameter required: cfg_path')
    alembic_cfg = Config(cfg_path)
    heads = get_heads_version(alembic_cfg)
    if len(heads) > 1:
        raise MigrationError(f'multiple heads detected! {heads}')
    current = get_current_version(alembic_cfg)
    if current == heads[0]:
        return
    logging.warning('alembic version is not up-to-date, \
        will retry migration...')
    retry_migration(alembic_cfg, 0, max_retry_count)



""" Request scope for FastAPI
"""

from typing import Dict, Callable

from styler_identity import Identity


class RequestScope(Identity):
    """ Carries information about the request scope
    """
    def __init__(
        self,
        authorization: str = None,
        accept_language: str = 'ja',
        trace: Dict = None,
    ) -> Callable:
        self.accept_language = accept_language
        self.trace = trace or {}
        token = get_token(authorization)
        super().__init__(token)

    def localization(self):
        # Also 'ja' when None is passed.
        return self.accept_language or 'ja'

    def trace_header(self):
        return self.trace

    @classmethod
    def from_request(cls, request):
        auth_header = request.headers.get('Authorization')
        accept_language = request.headers.get('Accept-Language', 'ja')
        trace = request.headers.get('trace_header', {})
        return cls(
            authorization=auth_header,
            accept_language=accept_language,
            trace=trace
        )


def get_token(authorization: str) -> str:
    """ Obtains the Access Token from the Authorization Header value
    """
    if not authorization:
        raise ValueError('Authorization header is expected')

    parts = authorization.split()

    if parts[0].lower() != 'bearer':
        raise ValueError('Authorization header must start with Bearer')
    elif len(parts) == 1:
        raise ValueError('Token not found')
    elif len(parts) > 2:
        raise ValueError('Authorization header must be Bearer token')

    token = parts[1]
    return token



""" Message handling module
"""

from typing import Any, Callable, Dict
import json
import logging

from styler_rest_framework.logging import error_reporting
from styler_rest_framework.config import defaults


class MessageRouter:
    """ Handles and routes messages to mapping handlers
    """

    def __init__(self, error_handler=None):
        self._routes = {}
        self.error_handler = error_handler or \
            error_reporting.google_error_reporting_handler(
                service=defaults.ERROR_HANDLER_SERVICE)

    def handle_message(self, message: Any) -> None:
        """ Route the incoming messages

            Args:
                message: message received from pull
        """
        logging.info('Received message: %s', message)

        message.ack()

        try:
            body = self._validate_envelope(message.data)

            message_name = body['name']
            if message_name in self._routes:
                self._routes[message_name](body['arg'])
            else:
                logging.info('Unknown message name: %s', message_name)
        except Exception:
            if self.error_handler:
                self.error_handler()
            logging.exception('Error when handling the message')

    def add_route(self,
                  message_name: str,
                  message_handler: Callable[[Any, Any], None]) -> None:
        """ Register a message route

            Args:
                message_name: expected message name.
                message_handler: function that receives a database connection
                    and the args received in the message.
        """
        self._routes[message_name] = message_handler

    def _validate_envelope(self, data: str) -> (bool, Dict):
        """ Validates the envelope of the data

            Args:
                data: the data received from the message
            Returns:
                The JSON parsed body data
        """
        body = json.loads(data)
        if 'name' not in body or 'arg' not in body:
            raise KeyError('Missing name or arg in the message body')
        return body



""" Defaults used by all modules
"""

import os


SERVICE_NAME = os.getenv('SERVICE_NAME') or 'svc'
NAMESPACE = os.getenv('NAMESPACE') or 'namespace'
VERSION = os.getenv('VERSION') or '0.1.0'

ERROR_HANDLER_SERVICE = f'{NAMESPACE}/{SERVICE_NAME}'



""" Tests for the shops publisher
"""

import json
from unittest.mock import Mock

from styler_rest_framework.pubsub import MessageRouter


class TestAddRoute:
    """ Tests for add_route
    """
    def test_add_route(self):
        router = MessageRouter()

        router.add_route('my_msg', Mock())

        assert 'my_msg' in router._routes


class TestHandleMessage:
    """ Tests for handle_message
    """
    def test_ack_message(self):
        message = Mock()
        router = MessageRouter()

        router.handle_message(message)

        message.ack.assert_called_once()

    def test_handle_errors_invalid_body(self):
        message = Mock()
        error_handler = Mock()
        router = MessageRouter(error_handler=error_handler)

        router.handle_message(message)

        error_handler.assert_called_once()

    def test_handle_errors_missing_parameters(self):
        message = Mock()
        message.data = json.dumps({'aaa': 'bbb'})
        error_handler = Mock()
        router = MessageRouter(error_handler=error_handler)

        router.handle_message(message)

        error_handler.assert_called_once()

    def test_handle_no_errors(self):
        message = Mock()
        message.data = json.dumps({'name': 'my_msg', 'arg': {'aaa': 'bbb'}})
        error_handler = Mock()
        router = MessageRouter(error_handler=error_handler)
        msg_handler = Mock()
        router.add_route('my_msg', msg_handler)

        router.handle_message(message)

        error_handler.assert_not_called()
        msg_handler.assert_called_once()

    def test_handle_no_route(self):
        message = Mock()
        message.data = json.dumps({'name': 'my_msg', 'arg': {'aaa': 'bbb'}})
        error_handler = Mock()
        router = MessageRouter(error_handler=error_handler)
        msg_handler = Mock()
        router.add_route('other', msg_handler)

        router.handle_message(message)

        error_handler.assert_not_called()
        msg_handler.assert_not_called()



""" Middleware to handle exceptions
"""
import logging

from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import JSONResponse


def add_exception_middleware(
        app: FastAPI,
        generic_message='An error has occurred',
        status_code=500,
        error_handler=None):
    """ Generate a middleware that logs unexpected exceptions
        and returns a JSON response.
        Exceptions of type HTTPException won't be handled as they should be
        expected exceptions.

        Args:
            app: FastAPI application
            generic_message: The message that will be send as an error
            status_code: The HTTP status code (default = 500)
            error_handler: Callable(request, exception)
    """
    @app.middleware('http')
    async def handle_exception(request: Request, call_next):
        try:
            response = await call_next(request)
            return response
        except HTTPException:
            raise
        except Exception as ex:
            message = str(ex)
            if error_handler:
                error_handler(request, ex)
            logging.exception('Error: %s', message)
            return JSONResponse(
                status_code=status_code,
                content={'error': generic_message},
            )



""" Define helper methods to reduce the boilerplate code for a server
"""

import logging

from styler_middleware import handle_exceptions, handle_invalid_json
from styler_rest_framework.config import defaults
from styler_rest_framework.logging import setup_logging
from styler_rest_framework.logging.error_reporting import (
    google_error_reporting_handler
)


def api_error_reporting_handler(service=None):  # pragma: no coverage
    try:
        service_name = service or defaults.ERROR_HANDLER_SERVICE
        from google.cloud import error_reporting
        handler = google_error_reporting_handler(service=service_name)

        def error_handler(request, exc):
            http_context = error_reporting.HTTPContext(
                method=request.method, url=request.path)
            handler(http_context=http_context)

        return error_handler
    except Exception:
        logging.warning('Could not find start api error reporting')
        return None


def add_middlewares(
            app,
            error_handler=None,
            service=None,
            handle_exceptions_args=None,
            handle_invalid_json_args=None
        ):
    """ Append default middlewares
    """
    handle_exceptions_args = handle_exceptions_args or {}
    handle_invalid_json_args = handle_invalid_json_args or {}
    error_handler = error_handler or api_error_reporting_handler(
        service=service)
    app.middlewares.extend([
        handle_exceptions(
            error_handler=error_handler,
            **handle_exceptions_args
        ),
        handle_invalid_json(**handle_invalid_json_args)
    ])


def set_logging(level=logging.INFO):  # pragma: no coverage
    setup_logging(level)


def default_configuration(app):  # pragma: no coverage
    set_logging()
    add_middlewares(app)



""" Tests for server defaults
"""

from unittest.mock import Mock

from styler_rest_framework.helpers import fastapi_defaults


class MockFastAPI:
    def __init__(self):
        self.middleware_func = None

    def middleware(self, middleware_type):
        def decorator(func):
            self.middleware_func = func
        return decorator


class TestAddMiddlewares:
    def test_add_middlewares(self):
        app = MockFastAPI()

        fastapi_defaults.add_middlewares(app)

        assert app.middleware_func is not None
        assert callable(app.middleware_func)

    def test_add_middlewares_with_custom_error_handler(self):
        app = MockFastAPI()
        error_handler = Mock()

        fastapi_defaults.add_middlewares(app, error_handler=error_handler)

        assert app.middleware_func is not None
        assert callable(app.middleware_func)



""" Custom implementation of encryption
"""

from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Hash import SHA256
from Crypto.Random import get_random_bytes


def encrypt(key: str, data: str):
    """ Encrypts the data string using a key

        The key will be used by PBKDF2 to derive a new key.
        This new key will be used in an AES cipher configured with EAX mode.

        Args:
            key: a secret string
            data: the data string to be encrypted
        Returns:
            A string containing the ciphertext, nonce, tag, and salt used in
            the encryption process.
    """
    if not key:
        raise ValueError('Key is missing')
    if not data:
        raise ValueError('Data is missing')
    salt = get_random_bytes(40)
    derived_key = PBKDF2(
        key, salt, 16, count=1000, hmac_hash_module=SHA256)
    cipher = AES.new(derived_key, AES.MODE_EAX)
    ciphertext, tag = cipher.encrypt_and_digest(data.encode())
    return ".".join([x.hex() for x in [ciphertext, cipher.nonce, tag, salt]])


def decrypt(key: bytes, data: str):
    """ Decrypts the data string using a key

        Args:
            key: a secret string
            data: the data string to be decrypted
        Returns:
            The decrypted string
    """
    if not key:
        raise ValueError('Key is missing')
    if not data:
        raise ValueError('Data is missing')
    try:
        ciphertext, nonce, tag, salt = [
            bytes.fromhex(v) for v in data.split('.')
        ]
        derived_key = PBKDF2(
            key, salt, 16, count=1000, hmac_hash_module=SHA256)
        cipher = AES.new(derived_key, AES.MODE_EAX, nonce=nonce)
        plaintext = cipher.decrypt_and_verify(ciphertext, tag)
        return plaintext.decode('utf-8')
    except Exception as ex:
        raise DecryptError('Error decrypting the data: %s', str(ex))


class DecryptError(Exception):
    """ Raised when an error happened when decrypting the data
    """



""" Tests for session module
"""
from unittest.mock import patch

from styler_rest_framework.datasource.sqlalchemy import session


class TestGetSession:
    """ Tests for get_session
    """
    @patch('styler_rest_framework.datasource.sqlalchemy.session.sessionmaker')
    @patch('styler_rest_framework.datasource.sqlalchemy.session.create_engine')
    def test_get_session(self, create_engine_mocked, sessionmaker_mocked):
        session_instance = session.get_session('uri')
        next(session_instance)

        create_engine_mocked.assert_called_once()
        sessionmaker_mocked.assert_called_once()



""" Tests for the exception middleware
"""
from unittest.mock import Mock, patch, AsyncMock

from styler_rest_framework.middlewares.fastapi import exception_middleware
from fastapi import HTTPException
from fastapi.responses import JSONResponse
import pytest


class MockFastAPI:
    def __init__(self):
        self.middleware_func = None

    def middleware(self, middleware_type):
        def decorator(func):
            self.middleware_func = func
        return decorator


class TestAddExceptionMiddleware:
    """ Tests for add_exception_middleware
    """
    def test_add_exception_middleware(self):
        app = MockFastAPI()

        exception_middleware.add_exception_middleware(app)

        assert app.middleware_func is not None
        assert callable(app.middleware_func)


class TestHandleException:
    """ Tests for handling exceptions
    """
    @patch('logging.exception')
    async def test_handle_exception(self, logging_mocked):
        app = MockFastAPI()
        error_handler = Mock()
        exception_middleware.add_exception_middleware(
            app, error_handler=error_handler)
        call_next = AsyncMock(side_effect=Exception('Error'))
        request = Mock()

        response = await app.middleware_func(request, call_next)

        error_handler.assert_called_once()
        logging_mocked.assert_called_once()
        assert isinstance(response, JSONResponse)

    @patch('logging.exception')
    async def test_handle_exception_without_handler(self, logging_mocked):
        app = MockFastAPI()
        exception_middleware.add_exception_middleware(app)
        call_next = AsyncMock(side_effect=Exception('Error'))
        request = Mock()

        response = await app.middleware_func(request, call_next)

        logging_mocked.assert_called_once()
        assert isinstance(response, JSONResponse)

    async def test_raise_http_exception(self):
        app = MockFastAPI()
        exception_middleware.add_exception_middleware(app)
        call_next = AsyncMock(side_effect=HTTPException(status_code=400))
        request = Mock()

        with pytest.raises(HTTPException):
            _ = await app.middleware_func(request, call_next)

    async def test_no_exception(self):
        app = MockFastAPI()
        exception_middleware.add_exception_middleware(app)
        call_next = AsyncMock(return_value=JSONResponse(status_code=200))
        request = Mock()

        response = await app.middleware_func(request, call_next)

        assert isinstance(response, JSONResponse)
        assert response.status_code == 200
